/*
  This is the ppl library. It provides a portable interface to usual OS features
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  The ppl library free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  The ppl library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with the ppl library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <ppl/ppl_dns.h>

#include <osip/port.h>
#include <osip/fifo.h>

fifo_t *dns_entries;		/* list of domain or FQDN strings to resolv */
smutex_t *m_dns_result;
ppl_dns_entry_t *dns_results;
ppl_dns_error_t *dns_errors;

static int
ppl_dns_lock_result_access (void)
{
  return smutex_lock (m_dns_result);
}

static int
ppl_dns_unlock_result_access (void)
{
  return smutex_unlock (m_dns_result);
}

#ifndef WIN32

static int
compare (const void *a, const void *b)
{
  ppl_dns_ip_t *aa, *bb;

  if (!a)
    return 1;
  if (!b)
    return -1;

  aa = (ppl_dns_ip_t *) * (int *) a;
  bb = (ppl_dns_ip_t *) * (int *) b;

  if (aa->pref > bb->pref)
    return 1;
  if (aa->pref < bb->pref)
    return -1;

  if (aa->rweight > bb->rweight)
    return -1;
  if (aa->rweight < bb->rweight)
    return 1;

  return 0;
}

#endif

PPL_DECLARE (char *) ppl_dns_get_next_query ()
{
  return fifo_get (dns_entries);
}

PPL_DECLARE (char *) ppl_dns_tryget_next_query ()
{
  return fifo_tryget (dns_entries);
}

PPL_DECLARE (ppl_status_t) ppl_dns_init ()
{
  dns_results = NULL;
  dns_errors = NULL;
  m_dns_result = smutex_init ();
  if (m_dns_result == NULL)
    return -1;
  dns_entries = (fifo_t *) smalloc (sizeof (fifo_t));
  if (dns_entries == NULL)
    {
      smutex_destroy (m_dns_result);
      sfree (m_dns_result);
      return -1;
    }
  fifo_init (dns_entries);
  return 0;
}

PPL_DECLARE (int) ppl_dns_ip_init (ppl_dns_ip_t ** ip)
{
  *ip = (ppl_dns_ip_t *) smalloc (sizeof (ppl_dns_ip_t));
  if (*ip == NULL)
    return -1;
  (*ip)->srv_ns_flag = PSP_NS_LOOKUP;	/* default is a NS lookup result */
  (*ip)->ttl = 60;
  (*ip)->pref = 0;
  (*ip)->name = NULL;
  (*ip)->port = 0;
  (*ip)->weight = 0;
  (*ip)->rweight = 0;
  (*ip)->next = NULL;
  (*ip)->parent = NULL;
  return 0;
}


PPL_DECLARE (int) ppl_dns_ip_free (ppl_dns_ip_t * ip)
{
  if (ip == NULL)
    return -1;
  sfree (ip->name);
  return 0;
}

PPL_DECLARE (int) ppl_dns_ip_clone (ppl_dns_ip_t * ip, ppl_dns_ip_t ** dest)
{
  ppl_dns_ip_t *tmp;
  int i;

  *dest = NULL;
  if (ip == NULL || ip->name == NULL)
    return -1;
  i = ppl_dns_ip_init (&tmp);
  if (i != 0)
    return -1;
  tmp->srv_ns_flag = ip->srv_ns_flag;
  tmp->ttl = ip->ttl;
  tmp->name = sgetcopy (ip->name);
  tmp->sin.sin_family = AF_INET;
  tmp->sin.sin_addr = ip->sin.sin_addr;
  tmp->sin.sin_port = ip->sin.sin_port;
  tmp->pref = ip->pref;
  tmp->port = ip->port;
  tmp->weight = ip->weight;
  tmp->rweight = ip->rweight;
  tmp->next = NULL;
  tmp->parent = NULL;
  *dest = tmp;
  return 0;
}


PPL_DECLARE (void) ppl_dns_entry_free (ppl_dns_entry_t * dns)
{
  ppl_dns_ip_t *ip;

  if (dns == NULL)
    return;
  sfree (dns->name);
  sfree (dns->protocol);
  for (ip = dns->dns_ips; ip != NULL; ip = dns->dns_ips)
    {
      REMOVE_ELEMENT (dns->dns_ips, ip);
      sfree (ip->name);
      sfree (ip);
    }
  return;
}


PPL_DECLARE (void) ppl_dns_entry_add_ref (ppl_dns_entry_t * dns)
{
  if (dns == NULL)
    return;
  dns->ref++;
}

PPL_DECLARE (void) ppl_dns_entry_remove_ref (ppl_dns_entry_t * dns)
{
  if (dns == NULL)
    return;
  dns->ref--;
}

PPL_DECLARE (ppl_status_t) ppl_dns_close ()
{
  ppl_dns_entry_t *dns;
  ppl_dns_error_t *err;

  for (dns = dns_results; dns != NULL; dns = dns_results)
    {
      REMOVE_ELEMENT (dns_results, dns);
      ppl_dns_entry_free (dns);
      sfree (dns);
    }
  for (err = dns_errors; err != NULL; err = dns_errors)
    {
      REMOVE_ELEMENT (dns_errors, err);
      sfree (err->domain);
      sfree (err);
    }
  smutex_destroy (m_dns_result);
  sfree (m_dns_result);

  fifo_free (dns_entries);
  sfree (dns_entries);

  /* endhostent(); */
  /* Not sure this is the correct method to call?
     Elements allocated by gethsotbyname_r still exists! */
  return 0;
}


PPL_DECLARE (void) ppl_dns_remove_entry (ppl_dns_entry_t * dns)
{
  ppl_dns_ip_t *d;

  if (dns->ref == 0)
    {
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_INFO4, NULL,
		   "DNS cache entry removed: %s \n", dns->name));
      REMOVE_ELEMENT (dns_results, dns);
      sfree (dns->name);
      sfree (dns->protocol);
      for (d = dns->dns_ips; d != NULL; d = dns->dns_ips)
	{
	  REMOVE_ELEMENT (dns->dns_ips, d);
	  sfree (d->name);
	  sfree (d);
	}
      sfree (dns);
    }
  else
    {
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_INFO4, NULL,
		   "DNS cache entry removed: %s \n", dns->name));
    }
}


/* I DON'T KNOW HOW TO SUPPORT SRV RECORD ON WINDOWS, TODO */
#ifdef WIN32
PPL_DECLARE (ppl_status_t)
ppl_dns_resolv (ppl_dns_ip_t ** dns_ips, querybuf * answer, int n)
{
  *dns_ips = NULL;
  return -1;
}

PPL_DECLARE (ppl_status_t)
ppl_dns_query (ppl_dns_entry_t ** dest, char *domain, char *protocol)
{
  *dest = NULL;
  return -1;
}
#else

PPL_DECLARE (ppl_status_t)
ppl_dns_resolv (ppl_dns_ip_t ** dns_ips, querybuf * answer, int n)
{
  ppl_dns_ip_t **array;
  int ancount, qdcount;		/* answer count and query count */
  HEADER *hp;			/* answer buffer header */
  char hostbuf[256];
  unsigned char *msg, *eom, *cp;	/* answer buffer positions */
  int dlen, type, aclass, pref, weight, port;
  long ttl;
  int answerno;

  *dns_ips = NULL;

  hp = (HEADER *) answer;
  qdcount = ntohs (hp->qdcount);
  ancount = ntohs (hp->ancount);

  msg = (unsigned char *) answer;
  eom = (unsigned char *) answer + n;
  cp = (unsigned char *) answer + sizeof (HEADER);

  while (qdcount-- > 0 && cp < eom)
    {
      n = dn_expand (msg, eom, cp, (char *) hostbuf, 256);
      if (n < 0)
	return -1;
      cp += n + QFIXEDSZ;
    }

  array = (ppl_dns_ip_t **) malloc (ancount * sizeof (ppl_dns_ip_t *));
  for (n = 0; n < ancount; n++)
    array[n] = NULL;
  answerno = 0;

  /* loop through the answer buffer and extract SRV records */
  while (ancount-- > 0 && cp < eom)
    {
      n = dn_expand (msg, eom, cp, (char *) hostbuf, 256);
      if (n < 0)
	{
	  for (n = 0; n < answerno; n++)
	    {
	      sfree (array[n]->name);
	      (void) free (array[n]);
	    }
	  (void) free (array);
	  return -1;
	}

      cp += n;

#if __OpenBSD__			/* !(__NAMESER > 19991005) some systems may need to check that? */
      type = _getshort (cp);
      cp += sizeof (u_short);
#else
      NS_GET16 (type, cp);
#endif

#if __OpenBSD__			/* !(__NAMESER > 19991005) some systems may need to check that? */
      aclass = _getshort (cp);
      cp += sizeof (u_short);
#else
      NS_GET16 (aclass, cp);
#endif

#if __OpenBSD__			/* !(__NAMESER > 19991005) some systems may need to check that? */
      ttl = _getlong (cp);
      cp += sizeof (u_long);
#else
      NS_GET32 (ttl, cp);
#endif

#if __OpenBSD__			/* !(__NAMESER > 19991005) some systems may need to check that? */
      dlen = _getshort (cp);
      cp += sizeof (u_short);
#else
      NS_GET16 (dlen, cp);
#endif

      if (type != T_SRV)
	{
	  cp += dlen;
	  continue;
	}
#if __OpenBSD__			/* !(__NAMESER > 19991005) some systems may need to check that? */
      pref = _getshort (cp);
      cp += sizeof (u_short);
#else
      NS_GET16 (pref, cp);
#endif

#if __OpenBSD__			/* !(__NAMESER > 19991005) some systems may need to check that? */
      weight = _getshort (cp);
      cp += sizeof (u_short);
#else
      NS_GET16 (weight, cp);
#endif

#if __OpenBSD__			/* !(__NAMESER > 19991005) some systems may need to check that? */
      port = _getshort (cp);
      cp += sizeof (u_short);
#else
      NS_GET16 (port, cp);
#endif

      n = dn_expand (msg, eom, cp, (char *) hostbuf, 256);
      if (n < 0)
	break;
      cp += n;

      array[answerno] = (ppl_dns_ip_t *) smalloc (sizeof (ppl_dns_ip_t));
      array[answerno]->srv_ns_flag = PSP_SRV_LOOKUP;
      array[answerno]->ttl = ttl;
      array[answerno]->pref = pref;
      array[answerno]->weight = weight;
      if (weight)
	array[answerno]->rweight = 1 + random () % (10000 * weight);
      else
	array[answerno]->rweight = 0;
      array[answerno]->port = port;
      array[answerno]->next = NULL;
      array[answerno]->name = sgetcopy ((char *) hostbuf);

      answerno++;
    }

  if (answerno == 0)
    {
      return -1;
    }

  qsort (array, answerno, sizeof (ppl_dns_ip_t *), compare);

  // Recreate a linked list from the sorted array...
  array[0]->parent = NULL;
  for (n = 0; n < answerno; n++)
    {
      if (n != 0)
	array[n]->parent = array[n - 1];
      array[n]->next = array[n + 1];
    }

  array[answerno - 1]->next = NULL;

  *dns_ips = array[0];
  (void) free (array);
  return 0;
}


PPL_DECLARE (ppl_status_t)
ppl_dns_query (ppl_dns_entry_t ** dest, char *domain, char *protocol)
{
  unsigned long int one_inet_addr;
  ppl_dns_entry_t *dns;
  ppl_dns_ip_t *ip;
  ppl_dns_ip_t *next;
  querybuf answer;		/* answer buffer from nameserver */
  int n;
  char *zone;

  *dest = NULL;

  if (!domain || !*domain || !protocol || !*protocol)
    return -1;

  zone = (char *) malloc (strlen (domain) + 5 + strlen (protocol) + 20);
  if (zone == NULL)
    return -1;

  *zone = '\0';

  strcat (zone, "_sip");
  strcat (zone, ".");
  strcat (zone, "_");
  strcat (zone, protocol);
  strcat (zone, ".");
  strcat (zone, domain);

  n =
    res_query (zone, C_IN, T_SRV, (unsigned char *) &answer, sizeof (answer));

  (void) free (zone);

  if (n < (int) sizeof (HEADER))
    {
      return -1;
    }
  /* valid answer received. */

  /* now we must resolv all IP address... */
  dns = (ppl_dns_entry_t *) smalloc (sizeof (ppl_dns_entry_t));

  dns->name = domain;
  dns->protocol = protocol;
  dns->date = ppl_time ();
  dns->dns_ips = NULL;
  dns->ref = 0;
  dns->next = NULL;
  dns->parent = NULL;

  n = ppl_dns_resolv (&(dns->dns_ips), &answer, n);
  if (n != 0)
    {
      ppl_dns_entry_free (dns);
      sfree (dns);
      return -1;
    }


  for (ip = dns->dns_ips; ip != NULL; ip = ip->next)
    {
      if ((int) (one_inet_addr = inet_addr (ip->name)) == -1)
	{
	  n = ppl_dns_gethostbyname (&(ip->sin), ip->name, ip->port);
	  if (n != PPL_SUCCESS)
	    {
	      sfree (ip->name);
	      ip->name = NULL;	/* mark for deletion */
	    }
	}
      else
	{
	  ip->sin.sin_port = htons ((short) ip->port);	/* to N. byte order */
	  ip->sin.sin_family = AF_INET;
	  ip->sin.sin_addr.s_addr = one_inet_addr;	/* already in N. byte order */
	}
    }

  /* delete ppl_dns_ip_t elements marked for deletion */

  for (ip = dns->dns_ips; ip != NULL; ip = next)
    {
      if (ip->name == NULL)
	{
	  next = ip->next;
	  REMOVE_ELEMENT (dns->dns_ips, ip);
	  sfree (ip);
	}
      else
	next = ip->next;
    }

  *dest = dns;

  return 0;
}

#endif

PPL_DECLARE (ppl_status_t) ppl_dns_error_add (char *address)
{
  ppl_dns_error_t *error;

  if (address == NULL)
    return -1;
  error = (ppl_dns_error_t *) smalloc (sizeof (ppl_dns_error_t));
  error->domain = address;
  error->next = NULL;
  error->parent = NULL;
  ppl_dns_lock_result_access ();
  ADD_ELEMENT (dns_errors, error);
  ppl_dns_unlock_result_access ();
  return 0;
}

PPL_DECLARE (ppl_status_t)
ppl_dns_get_error (ppl_dns_error_t ** dns_error, char *domain)
{
  ppl_dns_error_t *error;

  *dns_error = NULL;
  ppl_dns_lock_result_access ();
  for (error = dns_errors;
       error != NULL && 0 != strcmp (error->domain, domain);
       error = error->next)
    {
    }
  if (error == NULL)
    {
      ppl_dns_unlock_result_access ();
      return -1;
    }
  /* REMOVE_ELEMENT(dns_results, res); We could keep it for a while? */
  ppl_dns_unlock_result_access ();
  *dns_error = error;
  return 0;
}

/**
 * Add a result in the list of known result
 * @param dns the the dns result to add.
 */
PPL_DECLARE (ppl_status_t) ppl_dns_add_domain_result (ppl_dns_entry_t * dns)
{
  if (dns == NULL)
    return -1;
  ppl_dns_lock_result_access ();
  ADD_ELEMENT (dns_results, dns);
  ppl_dns_unlock_result_access ();
  return 0;
}

/**
 * A an entry in the list of known result
 * @param domain Domain to search
 */
PPL_DECLARE (ppl_status_t) ppl_dns_add_domain_query (char *domain)
{
  if (domain == NULL)
    return -1;
  fifo_add (dns_entries, domain);
  return 0;
}

PPL_DECLARE (ppl_status_t)
ppl_dns_get_result (ppl_dns_entry_t ** dns, char *domain)
{
  ppl_dns_entry_t *res;
  ppl_dns_entry_t *resnext;
  ppl_time_t now;

  now = ppl_time ();
  *dns = NULL;
  ppl_dns_lock_result_access ();

  res = dns_results;
  resnext = res;

  for (; res != NULL && 0 != strcmp (res->name, domain); res = resnext)
    {
      resnext = res->next;
      if (res->dns_ips != NULL)
	{
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_INFO4, NULL,
		       "time to live: %li, date: %i\n",
		       res->dns_ips->ttl, res->date));
	  if (now - res->date > res->dns_ips->ttl)	/* expired entry! */
	    {
	      OSIP_TRACE (osip_trace
			  (__FILE__, __LINE__, OSIP_INFO4, NULL,
			   "ENTRY REMOVED1\n"));
	      ppl_dns_remove_entry (res);
	    }
	}
    }
  if (res == NULL)
    {
      ppl_dns_unlock_result_access ();
      return -1;
    }

  if (res->dns_ips != NULL)
    {
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_INFO4, NULL,
		   "time to live: %li, date: %i\n",
		   res->dns_ips->ttl, res->date));
      if (now - res->date > res->dns_ips->ttl)	/* expired entry! */
	{
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_INFO4, NULL,
		       "ENTRY REMOVED2\n"));
	  ppl_dns_remove_entry (res);
	  res = NULL;
	  ppl_dns_unlock_result_access ();
	  return -1;
	}
    }

  /*  REMOVE_ELEMENT(dns_results, res); We could keep it for a while? */
  ppl_dns_unlock_result_access ();
  *dns = res;
  return 0;
}

#if defined(__OpenBSD__) || defined(WIN32)

PPL_DECLARE (ppl_status_t)
ppl_getaddrinfo (struct sockaddr_in * sin, char *hostname, char *service)
{
  unsigned long int one_inet_addr;
  struct addrinfo hints, *res0;
  int error;

  memset (&hints, 0, sizeof (hints));
  if ((int) (one_inet_addr = inet_addr (hostname)) == -1)
    hints.ai_flags = AI_CANONNAME;
  else
    hints.ai_flags = AI_NUMERICHOST;
  hints.ai_family = PF_INET;
  hints.ai_socktype = SOCK_DGRAM;
  hints.ai_protocol = IPPROTO_UDP;
  error = getaddrinfo (hostname, service, &hints, &res0);
  if (error || res0 == NULL)
    return -1;

  fprintf (stdout, "The canonnical name is: %s\n", res0->ai_canonname);
  fprintf (stdout, "The ai_addrlen is: %i\n", res0->ai_addrlen);

  memcpy (sin, res0->ai_addr, res0->ai_addrlen);
  freeaddrinfo (res0);

  return PPL_SUCCESS;
}

PPL_DECLARE (int)
ppl_dns_query_host (ppl_dns_entry_t ** dest, char *hostname, int port)
{
  struct sockaddr_in addr;
  ppl_dns_ip_t *dns_ip;
  ppl_dns_entry_t *dns;
  int my_error;

  *dest = NULL;

  my_error = ppl_getaddrinfo (&addr, hostname, "sip");
  if (my_error)
    return my_error;

  dns_ip = (ppl_dns_ip_t *) smalloc (sizeof (ppl_dns_ip_t));

  dns_ip->srv_ns_flag = PSP_NS_LOOKUP;
  dns_ip->ttl = 60;
  dns_ip->pref = 10;
  dns_ip->weight = 0;
  dns_ip->rweight = 0;
  dns_ip->port = port;
  dns_ip->next = NULL;
  dns_ip->parent = NULL;

  dns_ip->sin.sin_family = addr.sin_family;
  dns_ip->sin.sin_addr = addr.sin_addr;
  dns_ip->sin.sin_port = htons ((short) port);

  dns_ip->name = sgetcopy (inet_ntoa (addr.sin_addr));

  dns = (ppl_dns_entry_t *) smalloc (sizeof (ppl_dns_entry_t));
  if (dns == NULL)
    {
      /* MLEAK! I have to free first_ip and following dns_ip!! */
      sfree (dns_ip->name);
      sfree (dns_ip);
      return -1;
    }
  dns->name = hostname;
  dns->protocol = NULL;
  dns->date = ppl_time ();
  dns->dns_ips = dns_ip;
  dns->ref = 0;
  dns->next = NULL;
  dns->parent = NULL;

  *dest = dns;
  return 0;

}

#else

PPL_DECLARE (int)
ppl_dns_query_host (ppl_dns_entry_t ** dest, char *hostname, int port)
{
  struct hostent result_buffer;
  char tmp[GETHOSTBYNAME_BUFLEN];
  struct hostent *result;
  int my_error;

  ppl_dns_entry_t *dns;
  ppl_dns_ip_t *first_ip;
  ppl_dns_ip_t **array;
  char *hostbuf;
  int curaddr = 0;
  int n;

  *dest = NULL;

  /* Jay changed according to TRU64 plateform */
  /* #if defined(HAVE_FUNC_GETHOSTBYNAME_R_3) */
#if defined(_OSF_SOURCE)
  gethostbyname_r (hostname,	/* the FQDN */
		   &result_buffer,	/* the result buffer */
		   &result);
  /* #elif defined(HAVE_FUNC_GETHOSTBYNAME_R_5) */
#elif defined(__sun__)
  result = gethostbyname_r (hostname,	/* the FQDN */
			    &result_buffer,	/* the result buffer */
			    tmp, GETHOSTBYNAME_BUFLEN - 1, &my_error);
  /* #elif defined(HAVE_FUNC_GETHOSTBYNAME_R_6) */
#else /* linux */
  gethostbyname_r (hostname,	/* the FQDN */
		   &result_buffer,	/* the result buffer */
		   tmp, GETHOSTBYNAME_BUFLEN - 1, &result, &my_error);
#endif

  if (!result)
    return my_error;

  /* build a dns result entry and add it in the list */

  {
    int ancount = 1;

    while (result->h_addr_list[ancount])
      {
	++ancount;
      }

    array = (ppl_dns_ip_t **) malloc (ancount * sizeof (ppl_dns_ip_t *));
    for (n = 0; n < ancount; n++)
      array[n] = NULL;

    curaddr = 0;
    while (result->h_addr_list[curaddr])
      {
	hostbuf =
	  inet_ntoa (*(struct in_addr *) result->h_addr_list[curaddr]);
	array[curaddr] = (ppl_dns_ip_t *) smalloc (sizeof (ppl_dns_ip_t));
	array[curaddr]->srv_ns_flag = PSP_NS_LOOKUP;
	array[curaddr]->ttl = 60;
	array[curaddr]->pref = 10;
	array[curaddr]->weight = 0;
	if (array[curaddr]->weight)
	  array[curaddr]->rweight =
	    1 + random () % (10000 * array[curaddr]->weight);
	else
	  array[curaddr]->rweight = 0;
	array[curaddr]->port = port;
	array[curaddr]->next = NULL;
	array[curaddr]->parent = NULL;
	array[curaddr]->name = sgetcopy (hostbuf);
	array[curaddr]->sin.sin_family = AF_INET;
	array[curaddr]->sin.sin_addr =
	  *(struct in_addr *) result->h_addr_list[0];
	array[curaddr]->sin.sin_port = htons (port);

	++curaddr;
      }

    qsort (array, curaddr, sizeof (ppl_dns_ip_t *), compare);

    // Recreate a linked list from the sorted array...
    array[0]->parent = NULL;
    for (n = 0; n < curaddr; n++)
      {
	if (n != 0)
	  array[n]->parent = array[n - 1];
	array[n]->next = array[n + 1];
      }

    array[curaddr - 1]->next = NULL;

    first_ip = array[0];
    (void) free (array);

    dns = (ppl_dns_entry_t *) smalloc (sizeof (ppl_dns_entry_t));
    if (dns == NULL)
      {
	/* MLEAK! I have to free first_ip and following dns_ip!! */
	return -1;
      }
    dns->name = hostname;
    dns->protocol = NULL;
    dns->date = ppl_time ();
    dns->dns_ips = first_ip;
    dns->ref = 0;
    dns->next = NULL;
    dns->parent = NULL;

    *dest = dns;
  }

  {
    /* this method also return the first IP address in the sin structure. */
#ifdef HAVE_ARPA_INET_H
    OSIP_TRACE (osip_trace ((__FILE__, __LINE__, OSIP_INFO4, NULL,
			     "IP addresses used: %s\n",
			     inet_ntoa (*(struct in_addr *) result->
					h_addr_list[0])));
		curaddr = 0;
		++curaddr; while (result->h_addr_list[curaddr])
		{
		OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO4, NULL,
					"IP addresses to be tried later: %s\n",
					inet_ntoa (*(struct in_addr *)
						   result->
						   h_addr_list[curaddr])));
		++curaddr;}
#endif
		}
		return PPL_SUCCESS;}

#endif


#if defined(__OpenBSD__) || defined(WIN32)
/* use getaddrinfo instead. */
/* I find that the BSD getaddrinfo is reentrant, but the manual say it's not? */
		PPL_DECLARE (int)
		ppl_dns_gethostbyname (struct sockaddr_in *sin,
				       char *hostname, int port)
		{
		struct addrinfo hints, *res0;
		int error;
		memset (&hints, 0, sizeof (hints));
		hints.ai_flags = AI_CANONNAME;
		hints.ai_family = PF_INET;
		hints.ai_socktype = SOCK_DGRAM;
		hints.ai_protocol = IPPROTO_UDP;
		error = getaddrinfo (hostname, "sip", &hints, &res0);
		if (error || res0 == NULL)
		return -1;
		fprintf (stdout, "The canonnical name is: %s\n",
			 res0->ai_canonname);
		fprintf (stdout, "The ai_addrlen is: %i\n", res0->ai_addrlen);
		memcpy (sin, res0->ai_addr, res0->ai_addrlen);
		freeaddrinfo (res0); return PPL_SUCCESS;}


#else

		PPL_DECLARE (int)
		ppl_dns_gethostbyname (struct sockaddr_in *sin,
				       char *hostname, int port)
		{
		struct hostent result_buffer;
		char tmp[GETHOSTBYNAME_BUFLEN];
		struct hostent * result; int my_error;
/* Jay changed according to TRU64 plateform */
		/* #if defined(HAVE_FUNC_GETHOSTBYNAME_R_3) */
#if defined(_OSF_SOURCE)||defined(HAVE_FUNC_GETHOSTBYNAME_R_3)
		gethostbyname_r (hostname,	/* the FQDN */
				 &result_buffer,	/* the result buffer */
				 &result);
		/* #elif defined(HAVE_FUNC_GETHOSTBYNAME_R_5) */
#elif defined(__sun__)||defined(HAVE_FUNC_GETHOSTBYNAME_R_5)
		result = gethostbyname_r (hostname,
					  &result_buffer,
					  tmp, GETHOSTBYNAME_BUFLEN - 1,
					  &my_error);
#elif defined(HAVE_FUNC_GETHOSTBYNAME_R_6)
		gethostbyname_r (hostname,	/* the FQDN */
				 &result_buffer,	/* the result buffer */
				 tmp, GETHOSTBYNAME_BUFLEN - 1, &result,
				 &my_error);
#else
#error A reentrant method for gethostbyname_r() is not found
#endif
		if (!result) return my_error;
		{
		int curaddr = 0;
		sin->sin_family = AF_INET;
		sin->sin_addr = *(struct in_addr *) result->h_addr_list[0];
		sin->sin_port = htons (port);
#ifdef HAVE_ARPA_INET_H
		OSIP_TRACE (osip_trace ((__FILE__, __LINE__, OSIP_INFO4, NULL,
					 "IP addresses used: %s\n",
					 inet_ntoa (*(struct in_addr *)
						    result->h_addr_list[0])));
			    ++curaddr; while (result->h_addr_list[curaddr])
			    {
			    OSIP_TRACE (osip_trace
					(__FILE__, __LINE__, OSIP_INFO4, NULL,
					 "IP addresses to be tried later: %s\n",
					 inet_ntoa (*(struct in_addr *)
						    result->
						    h_addr_list[curaddr])));
			    ++curaddr;}
#endif
			    }
			    return PPL_SUCCESS;}

#endif
